# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Dispatcher forwards requests to Services."""

from logging import getLogger
from typing import List, Type, cast

from mitmproxy.http import HTTPFlow, HTTPResponse

from .errors import DoNotIntercept, NoServicesForRequest
from .management_service import ManagementService
from .services import ArxivService, BaseService
from .translator import dict_to_response, request_to_dict
from .whitelist_service import WhitelistService

logger = getLogger(__name__)


class Dispatcher:
    SERVICE_LIST: List[Type[BaseService]] = [
        ArxivService,
        WhitelistService,
    ]

    def __init__(self) -> None:
        self.services = [service_class() for service_class in self.SERVICE_LIST]
        mgmt_service = ManagementService(self.services)
        self.services = [cast(BaseService, mgmt_service)] + self.services

    def process_request(self, request: dict) -> dict:
        """Perform operations and give response."""
        for service in self.services:
            if service.handles_request(request):
                return service.process_request(request)
        raise NoServicesForRequest(request)

    def request(self, flow: HTTPFlow):
        """MITMProxy addon event interface for outgoing request."""
        try:
            request = request_to_dict(flow.request)
            response = dict_to_response(self.process_request(request))
            flow.response = response
        except DoNotIntercept as e:
            # Let the request pass through, by not interrupting the flow, but log it
            logger.warning(str(e))
        except Exception as e:
            flow.response = HTTPResponse.make(
                status_code=getattr(e, 'http_status_code', 500),
                content=str(e),
                headers={'Content-Type': 'text/plain'}
            )



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Tools to convert between responses and request formats."""

from cgi import parse_header
from typing import Dict, List

from mitmproxy.http import HTTPRequest, HTTPResponse
from mitmproxy.net.http.headers import Headers


def request_to_dict(request: HTTPRequest) -> dict:
    encoding = encoding_by_header('Accept', request.headers)

    return {
        'method': request.method,
        'uri': request.url,
        'body': request.content.decode(encoding),
        'headers': {
            k: request.headers.get_all(k)
            for k in request.headers.keys()
        },
    }


def dict_to_response(response: dict) -> HTTPResponse:
    headers = dict_to_headers(response['headers'])
    encoding = encoding_by_header('content-type', headers)

    content = response['body']
    if isinstance(content, str):
        content = content.encode(encoding)

    return HTTPResponse(
        http_version='HTTP/1.1',
        status_code=response['status']['code'],
        reason=response['status']['message'].encode('ascii'),
        headers=headers,
        content=content,
    )


def headers_to_dict(headers: Headers) -> dict:
    header_dict: Dict[str, List[str]] = {}

    for key, value in headers.fields:
        key, value = key.decode('ascii'), value.decode('ascii')

        if key in header_dict:
            header_dict[key].append(value)
        else:
            header_dict[key] = [value]

    return header_dict


def dict_to_headers(headers: dict) -> Headers:
    fields = []

    for key, values in headers.items():
        for value in values:
            fields.append(
                (key.encode('ascii'), value.encode('ascii'))
            )

    return Headers(fields=fields)


def encoding_by_header(header_name: str, headers: Headers) -> str:
    """Extract charset param from Content-Type or Accept headers"""
    try:
        content_type = headers[header_name]
        _, params = parse_header(content_type)
        return params['charset']
    except KeyError:
        return 'utf-8'



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""INSPIRE-MITMProxy Errors"""


class MITMProxyHTTPError(Exception):
    http_status_code = 500


class NoServicesForRequest(MITMProxyHTTPError):
    def __init__(self, request: dict) -> None:
        self.http_status_code = 501
        message = f"None of the registered services can handle this request: {request}"
        super().__init__(message)


class RequestNotHandledInService(MITMProxyHTTPError):
    def __init__(self, service_name: str, request: dict) -> None:
        self.http_status_code = 501
        message = f"{service_name} can't handle the request {request}"
        super().__init__(message)


class InvalidRequest(MITMProxyHTTPError):
    def __init__(self, service_name: str, request: dict) -> None:
        self.http_status_code = 400
        message = f"Invalid request {request} for service {service_name}"
        super().__init__(message)


class DoNotIntercept(Exception):
    def __init__(self, service_name: str, request: dict) -> None:
        message = f"Allow request {request} in {service_name} to pass through to the outside"
        super().__init__(message)


class NoMatchingRecording(MITMProxyHTTPError):
    def __init__(self, service_name: str, request: dict) -> None:
        self.http_status_code = 501
        message = f"Service {service_name} cannot handle this request: {request}"
        super().__init__(message)


class ScenarioNotFound(MITMProxyHTTPError):
    def __init__(self, service_name: str, scenario: str) -> None:
        self.http_status_code = 501
        message = f"Scenario {scenario} not found in service {service_name}"
        super().__init__(message)


class ScenarioUndefined(MITMProxyHTTPError):
    def __init__(self, service_name: str) -> None:
        self.http_status_code = 501
        message = f"Scenario not set in service {service_name}"
        super().__init__(message)



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Base for fake services."""

from os import environ
from pathlib import Path
from typing import List, Optional
from urllib.parse import splitport  # type: ignore
from urllib.parse import urlparse

from yaml import load as yaml_load

from ..errors import NoMatchingRecording, ScenarioNotFound, ScenarioUndefined


class BaseService:
    """Mocked service base."""
    SERVICE_HOSTS: List[str] = []
    active_scenario: Optional[str] = None

    @property
    def name(self):
        return type(self).__name__

    def handles_request(self, request: dict) -> bool:
        """Can this service handle the request?"""
        try:
            host = splitport(request['headers']['Host'][0])[0]
        except KeyError:
            host = urlparse(request['uri']).hostname

        return host in self.SERVICE_HOSTS

    def process_request(self, request: dict) -> dict:
        """Perform operations and give response."""
        for recorded_response in self.get_responses_for_active_scenario():
            if self.match_request(request, recorded_response['request']):
                return recorded_response['response']

        raise NoMatchingRecording(self.name, request)

    def match_request(self, incoming_request: dict, recorded_request: dict) -> bool:
        parsed_incoming_uri = urlparse(incoming_request['uri'])
        parsed_recorded_uri = urlparse(recorded_request['uri'])

        return (
            incoming_request['method'] == recorded_request['method']
            and parsed_incoming_uri == parsed_recorded_uri
            and (incoming_request['body'] or None) == (recorded_request['body'] or None)
        )

    def get_responses_for_active_scenario(self) -> List[dict]:
        """Get a list of scenarios"""
        if not self.active_scenario:
            raise ScenarioUndefined(self.name)

        scenarios_path = Path(environ.get('SCENARIOS_PATH', './scenarios/'))
        responses_dir = scenarios_path / self.active_scenario / self.name

        if not responses_dir.exists():
            raise ScenarioNotFound(self.name, self.active_scenario)

        responses = [
            yaml_load(response.read_text()) for response in responses_dir.iterdir()
            if response.is_file() and response.suffix == '.yaml'
        ]

        return responses



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""INSPIRE MITMProxy for E2E Tests."""

from typing import List

from setuptools import setup

readme = open('README.rst').read()

setup_requires = [
    'autosemver~=0.0,>=0.5.3',
]

install_requires = [
    'autosemver~=0.0,>=0.5.3',
    'mitmproxy~=3.0,>=3.0.4',
    'pathlib~=1.0,>=1.0.1',
    'pyyaml~=3.0,>=3.12',
]

tests_require = [
    'flake8~=3.0,>=3.5.0',
    'isort~=4.0,>=4.3.4',
    'mypy~=0.0,>=0.590',
    'mock~=2.0,>=2.0.0',
    'pytest~=3.0,>=3.5.0',
    'pytest-cov~=2.0,>=2.5.1',
]

docs_require: List[str] = []

extras_require = {
    'docs': docs_require,
    'tests': tests_require,
}

extras_require['all'] = []
for name, reqs in extras_require.items():
    extras_require['all'].extend(reqs)

setup(
    description=__doc__,
    long_description=readme,
    url='https://github.com/inspirehep/inspire-mitmproxy',
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    extras_require=extras_require,
    autosemver={'bugtracker_url': '/issues'}
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Tool to convert VCRpy casettes to request-response pair files.

Usage: vcr_convert.py [--help] CASETTE_1 [CASETTE_2 ...] DESTINATION_DIR

This will convert an arbitrary number of VCRpy casettes to files with
a single pair of request and response each, structured in folders according
to the pattern: DESTINATION_DIR/<scenario>/<host>/<index in casette>.yaml
"""

import sys
from pathlib import Path
from urllib.parse import urlparse

import yaml


def service_name_for_interaction(interaction: dict) -> str:
    parsed_url = urlparse(interaction['request']['uri'])
    return parsed_url.netloc


if len(sys.argv) < 3 or sys.argv[1] == '--help':
    print(__doc__)
    exit(1)

destination = Path(sys.argv[-1])

for cassette_file in sys.argv[1:-1]:
    cassette_file = Path(cassette_file)
    cassette = yaml.load(cassette_file.read_text())

    for idx, interaction in enumerate(cassette['interactions']):
        output_dir = destination / cassette_file.stem / service_name_for_interaction(interaction)
        output_dir.mkdir(parents=True, exist_ok=True)
        output = output_dir / f'{cassette_file.stem}.{idx}.yaml'
        print(f'Creating {output}')

        # Remove VCR's {'string': ...} wrapper:
        if interaction['response']['body'] is not None:
            interaction['response']['body'] = interaction['response']['body']['string']

        output.write_text(yaml.dump(interaction))



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Service used to orchestrate fake services."""

from json import JSONDecodeError
from json import dumps as json_dumps
from json import loads as json_loads
from os import environ, getcwd
from pathlib import Path
from typing import Dict, List, Optional, Union, cast
from urllib.parse import urlparse

from autosemver import get_current_version
from mitmproxy.net.http.status_codes import RESPONSES

from .errors import InvalidRequest, RequestNotHandledInService
from .services import BaseService


class ManagementService(BaseService):
    SERVICE_HOSTS = ['mitm-manager.local']

    def __init__(self, services: List[BaseService]) -> None:
        self.services = services
        self.config = {
            'active_scenario': None,
        }

    def process_request(self, request: dict) -> dict:
        parsed_url = urlparse(request['uri'])
        path = parsed_url.path
        method = request['method']

        if path == '/services' and method == 'GET':
            return self.build_response(200, self.get_services())
        elif path == '/scenarios' and method == 'GET':
            return self.build_response(200, self.get_scenarios())
        elif path == '/config' and method == 'GET':
            return self.build_response(200, self.get_config())
        elif path == '/config' and method == 'PUT':
            return self.build_response(204, self.put_config(request))
        elif path == '/config' and method == 'POST':
            return self.build_response(201, self.post_config(request))

        raise RequestNotHandledInService(self.name, request)

    def get_services(self) -> dict:
        return {
            idx: {
                'class': service.name,
                'service_hosts': service.SERVICE_HOSTS,
            }
            for idx, service in enumerate([cast(BaseService, self)] + self.services)
        }

    def get_scenarios(self) -> dict:
        path = Path(environ.get('SCENARIOS_PATH', './scenarios/'))
        response: Dict[str, Dict[str, Dict[str, List[str]]]] = {
            scenario.name: {
                'responses': {
                    service.name: [
                        response_file.name for response_file in sorted(service.iterdir())
                        if response_file.is_file() and response_file.suffix == '.yaml'
                    ]
                    for service in scenario.iterdir() if service.is_dir()
                }
            }
            for scenario in path.iterdir() if scenario.is_dir()
        }

        return response

    def get_config(self) -> dict:
        return self.config

    def put_config(self, request: dict):
        try:
            config_update = json_loads(request['body'])
            self.config.update(config_update)
            self.config_propagate()
        except (JSONDecodeError, ValueError):
            raise InvalidRequest(self.name, request)

    def post_config(self, request: dict):
        try:
            self.config = json_loads(request['body'])
            self.config_propagate()
        except (JSONDecodeError):
            raise InvalidRequest(self.name, request)

    def build_response(self, code: int, json_message: Optional[Union[dict, list]]) -> dict:
        if json_message:
            body = json_dumps(json_message, indent=2)
        else:
            body = ''

        return {
            'status': {
                'message': RESPONSES[code],
                'code': code,
            },
            'body': body,
            'headers': {
                'Content-Type': ['application/json'],
                'Server': ['inspire-mitmproxy/' + get_current_version(getcwd())]
            }
        }

    def config_propagate(self):
        """On change of config, propagate relevant information to services."""
        for service in self.services:
            service.active_scenario = self.config.get('active_scenario')
